"""

.. image:: https://img.shields.io/pypi/pyversions/mypythontools.svg
    :target: https://pypi.python.org/pypi/mypythontools/
    :alt: Python versions

.. image:: https://badge.fury.io/py/mypythontools.svg
    :target: https://badge.fury.io/py/mypythontools
    :alt: PyPI version

.. image:: https://img.shields.io/lgtm/grade/python/github/Malachov/mypythontools.svg
    :target: https://lgtm.com/projects/g/Malachov/mypythontools/context:python
    :alt: Language grade: Python

.. image:: https://readthedocs.org/projects/mypythontools/badge/?version=latest
    :target: https://mypythontools.readthedocs.io/en/latest/?badge=latest
    :alt: Documentation Status

.. image:: https://img.shields.io/badge/License-MIT-yellow.svg
    :target: https://opensource.org/licenses/MIT
    :alt: License: MIT


Some tools/functions/snippets used across projects.

Usually used from IDE. Used paths are infered and things like building the
application with pyinstaller, incrementing version, generating rst files for sphinx docs,
pushing to github or deploying to Pypi is matter of calling one function
or clicking one button (e.g. Vs code task).

Many projects - one codebase.

If you are not sure whether structure of your app will work with this code, there is python starter repo
on https://github.com/Malachov/my-python-starter

Paths are infered, but if you have atypical structure or have more projects in cwd, use `mypythontools.misc.set_paths()`.

Links
=====

Official documentation - https://mypythontools.readthedocs.io/
Official repo - https://github.com/Malachov/mypythontools

Installation
============

Python >=3.6 (Python 2 is not supported).

Install just with::

    pip install mypythontools

"""

from . import utils
from . import build
from . import deploy
from . import misc
from . import pyvueeel

__version__ = "0.0.34"

__author__ = "Daniel Malachov"
__license__ = "MIT"
__email__ = "malachovd@seznam.cz"

__all__ = ["utils", "build", "deploy", "misc", "pyvueeel"]



"""
Allow to deploy project.
Possible destinations: PYPI.
"""
import subprocess
import os
import mylogging
import shutil

from . import misc


def deploy_to_pypi(setup_path=None):
    """Publish python library to Pypi. Username and password are set
    with env vars `TWINE_USERNAME` and `TWINE_PASSWORD`.

    Args:
        setup_path((str, pathlib.Path)): Function suppose, that there is a setup.py in cwd. If not, pass path to setup.py.
    """
    try:
        os.environ['TWINE_USERNAME']
        os.environ['TWINE_PASSWORD']
    except KeyError:
        raise KeyError(mylogging.return_str('Setup env vars TWINE_USERNAME and TWINE_PASSWORD to use deploy.'))

    setup_path = misc.root_path if not setup_path else setup_path

    setup_py_path = setup_path / 'setup.py'

    if not setup_py_path.exists():
        setup_py_path = misc.find_path('setup.py', exclude=['node_modules'])
        setup_path = setup_py_path.parent

        if not (setup_path / 'setup.py').exists():
            raise FileNotFoundError(mylogging.return_str("Setup.py file not found. Setup `setup_path` param."))

    try:
        shutil.rmtree(setup_path / 'dist')
        shutil.rmtree(setup_path / 'build')

    except Exception:
        pass

    subprocess.run(['python', 'setup.py', 'sdist', 'bdist_wheel'], cwd=setup_path, shell=True, check=True)

    subprocess.run(['twine', 'upload', '-u', os.environ['TWINE_USERNAME'], '-p', os.environ['TWINE_PASSWORD'], 'dist/*'], cwd=setup_path, shell=True, check=True)

    shutil.rmtree(setup_path / 'dist')
    shutil.rmtree(setup_path / 'build')



"""
Common functions for Python / Vue / Eel project.

It converts json to correct python format dict or transform data
into form for vue table or plot.

Run `mypythontools.pyvueeel.help_starter_pack_vue_app()` for tutorial how to create such an app.
"""

import os
import sys
from pathlib import Path
import warnings
import numpy as np

import mylogging

from . import misc

# Lazy load
# import pandas as pd
# import EelForkExcludeFiles as eel


expose_error_callback = None


def run_gui(
    devel=None, is_multiprocessing=False, log_file_path=None, builded_gui_path=None
):
    """Function that init and run `eel` project.
    It will autosetup chrome mode (if installed chrome or chromium, open separate window with
    no url bar, no bookmarks etc...) if chrome is not installed, it open microsoft Edge (by default
    on windows).

    In devel mode, app is connected on live vue server. Serve your web application and debug python file
    caling this function. Debugger should correctly stop at breakpoints.

    Args:
        devel((bool, None), optional): If None, detected. Can be overwritten. Devel 0 run static assets, 1 run Vue server on localhost. Defaults to None.
        is_multiprocessing (bool, optional): If using multiprocessing in some library, set up to True. Defaults to False.
        log_file_path ((str, Path, None)), optional): If not exist, it will create, if exist, it will append,
            if None, log to relative log.log and only if in production mode.
        builded_gui_path ((str, Path, None)), optional): Where the web asset is. Only if debug is 0 but not run with pyinstaller. If None, it's automatically find (but is slower then). Defaults to None.
    """

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", module="EelForkExcludeFiles", category=ResourceWarning
        )

        import EelForkExcludeFiles as eel

    try:
        if devel is None:
            # env var MY_PYTHON_VUE_ENVIRONMENT is configured and added with pyinstaller automatically in build module
            devel = (
                False
                if os.environ.get("MY_PYTHON_VUE_ENVIRONMENT") == "production"
                else True
            )

        # Whether run is from .exe or from python
        is_builded = True if getattr(sys, "frozen", False) else False

        if log_file_path:
            log_file = log_file_path
        else:
            log_file = "log.log" if is_builded else None

        mylogging.config.TO_FILE = log_file

        if is_builded:
            # gui folder is created with pyinstaller in build
            gui_path = Path(sys._MEIPASS) / "gui"
        else:
            if devel:
                misc.set_root()
                gui_path = (
                    misc.find_path(
                        "index.html", exclude=["node_modules", "build"]
                    ).parents[1]
                    / "src"
                )
            else:
                if builded_gui_path:
                    gui_path = Path(builded_gui_path)
                else:
                    misc.set_root()
                    gui_path = misc.find_path(
                        "index.html", exclude=["public", "node_modules", "build"]
                    ).parent

        if not gui_path.exists():
            raise FileNotFoundError(
                "Web files not found, setup gui_path (where builded index.html is)."
            )

        if devel:
            directory = gui_path
            app = None
            page = {"port": 8080}
            port = 8686
            init_files = [".vue", ".js", ".html"]

            def close_callback(page, sockets):
                pass

        else:
            directory = gui_path
            close_callback = None
            app = "chrome"
            page = "index.html"
            port = 0
            init_files = [".js", ".html"]

        eel.init(directory.as_posix(), init_files, exlcude_patterns=["chunk-vendors"])

        if is_multiprocessing:
            from multiprocessing import freeze_support

            freeze_support()

        mylogging.info("Py side started")

        eel.start(
            page,
            mode=app,
            cmdline_args=["--disable-features=TranslateUI"],
            close_callback=close_callback,
            host="localhost",
            port=port,
            disable_cache=True,
        ),

    except OSError:
        eel.start(
            page,
            mode="edge",
            host="localhost",
            close_callback=close_callback,
            port=port,
            disable_cache=True,
        ),

    except Exception:
        mylogging.traceback("Py side terminated...")


def help_starter_pack_vue_app():
    """
    Tutorial how to build app with python, Vue and eel.
    Print help on build Vue part with CLI (many options, rather not copying files from other project).
    Then will help with basic structre of components, then show copypasters to other places.

    Structure

    - myproject
        - gui
            - generated with Vue CLI
        - app.py

    ############
    ### app.py
    ###########

    from mypythontools import pyvueeel
    from mypythontools.pyvueeel import expose

    # Expose python functions to Js with decorator
    @expose
        def load_data(settings):
            # You can return dict - will be object in js
            # You can return list - will be an array in js

            return {'Hello': 1}

    # Call function from JS
    pyvueeel.eel.myfunction()

    # End of file
    if __name__ == '__main__':
        pyvueeel.run_gui()

    #########
    ### gui
    ########

    Generate gui folder with Vue CLI

    ```console
    npm install -g @vue/cli
    vue create gui
    ```

    Goto folder and optionally

    ```console
    vue add vuex
    vue add vuetify
    vue add router
    ```

    #############
    ### main.js
    ############

    ```js
    if (process.env.NODE_ENV == 'development') {

    try {
        window.eel.set_host("ws://localhost:8686");

    } catch (error) {
        document.getElementById('app').innerHTML = 'Py side is not running. Start app.py with debugger.'
        console.error(error);
    }

    Vue.config.productionTip = true
    Vue.config.devtools = true
    } else {
    Vue.config.productionTip = false
    Vue.config.devtools = false
    }

    // You can expose function to be callable from python. Import and then
    // window.eel.expose(function_name, 'function_name')

    ##########
    ### .env
    #########

    create empty files .env.development and add `VUE_APP_EEL=http://localhost:8686/eel.js`

    create empty .env.production and add `VUE_APP_EEL=eel.js`

    #################
    ### index.html
    ###############

    In public folder

    ```html
    <script type="text/javascript" src="<%= VUE_APP_EEL %>"></script>
    ```

    ###################
    ### vue.config.js
    #################

    ```js
    let devtool_mode
    if (process.env.NODE_ENV === 'development') {
    devtool_mode = 'source-map';
    } else {
    devtool_mode = false;
    }

    module.exports = {
    outputDir: "web_builded",
    transpileDependencies: [
        "vuetify"
    ],
    productionSourceMap: process.env.NODE_ENV != 'production',

    configureWebpack: {
        devtool: devtool_mode,
    }
    }
    ```

    #################
    ### Tips, trips
    ################

    # VS Code plugins for developing
    - npm
    - vetur
    - Vue VSCode Snippets
    - vuetify-vscode
    """

    print(help_starter_pack_vue_app.__doc__)


def expose(callback_function):
    """Wrap eel expose with try catch block and adding exception callback function
    (for printing error to frontend usually).

    Args:
        callback_function (function): Function that will be called if exposed function fails on some error.
    """

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", module="EelForkExcludeFiles", category=ResourceWarning
        )

        import EelForkExcludeFiles as eel

    def inner(*args, **kargs):
        try:
            return callback_function(*args, **kargs)

        except Exception:
            mylogging.traceback(
                f"Unexpected error in function `{callback_function.__name__}`"
            )
            if expose_error_callback:
                expose_error_callback()

    eel._expose(callback_function.__name__, inner)


def json_to_py(json):
    """Take json / json from JS side and eval it from strings.
    If string to string, if float to float, if object then to dict.

    When to use? - If sending object as parameter in function.

    Args:
        json (dict): Object from JS.

    Returns:
        dict: Python dictionary with correct types.
    """

    evaluated = {}
    for i, j in json.items():

        if j == "true":
            j = True
        if j == "false":
            j = False

        try:
            evaluated[i] = eval(j)
        except Exception:
            evaluated[i] = j

    return evaluated


def to_vue_plotly(data, names=None):
    """Takes data (dataframe or numpy array) and transforms it to form, that vue-plotly understand.

    Args:
        data ((np.array, pd.DataFrame)): Plotted data.
        names (list, optional): If using array, you can define names. Defaults to None.

    Returns:
        dict: Data in form for plotting in frontend.
    """

    import pandas as pd

    if isinstance(data, (np.ndarray, np.generic)):
        data = pd.DataFrame(data, columns=names)

    data = pd.DataFrame(data)

    numeric_data = data.select_dtypes(include="number").round(decimals=3)
    numeric_data = numeric_data.where(np.isfinite(numeric_data), None)
    # numeric_data = add_none_to_gaps(numeric_data)

    return {
        "x_axis": numeric_data.index.to_list(),
        "y_axis": numeric_data.values.T.tolist(),
        "names": numeric_data.columns.values.tolist(),
    }


def to_table(df):
    """Takes data (dataframe or numpy array) and transforms it to form, that vue-plotly library understands.

    Args:
        df (pd.DataFrame): Data in table form

    Returns:
        dict: Data in form for create table.
    """
    data = df.copy()
    data = data.round(decimals=3)

    # Numpy nan cannot be send to json - replace with None
    data = data.where(~data.isin([np.nan, np.inf, -np.inf]), None)

    headers = [{"text": i, "value": i, "sortable": True} for i in data.columns]

    return {"table": data.to_dict("records"), "headers": headers}



"""
This module can be used for example in running deploy pipelines or githooks
(some code automatically executed before commit). This module can run the tests,
edit library version, generate rst files for docs, push to git or deploy app to pypi.

All of that can be done with one function call - with `push_pipeline` function that
run other functions, or you can use functions separately. If you are using other
function then `push_pipeline`, you need to call `mypythontools.misc.set_paths()` first.


Examples:
=========

    **VS Code Task example**

    You can push changes with single click with all the hooks displaying results in
    your terminal. All params changing every push (like git message or tag) can
    be configured on the beginning and therefore you don't need to wait for test finish.
    Default values can be also used, so in small starting projects, push is actually very fast.

    Create folder utils, create push_script.py inside, add

    >>> import mypythontools
    ...
    >>> if __name__ == "__main__":
    >>>     mypythontools.utils.push_pipeline(deploy=True)  # With all the params you need.

    Then just add this task to global tasks.json::

        {
            "version": "2.0.0",
            "tasks": [
                {
                    "label": "Hooks & push & deploy",
                    "type": "shell",
                    "command": "python",
                    "args": [
                        "${workspaceFolder}/utils/push_script.py",
                        "--commit_message",
                        "${input:git-message}",
                        "--tag",
                        "${input:git-tag}",
                        "--tag_mesage",
                        "${input:git-tag-message}"
                    ],
                    "presentation": {
                        "reveal": "always",
                        "panel": "new"
                    }
                }
            ],
            "inputs": [
                {
                    "type": "promptString",
                    "id": "git-message",
                    "description": "Git message for commit.",
                    "default": "New commit"
                },
                {
                    "type": "promptString",
                    "id": "git-tag",
                    "description": "Git tag.",
                    "default": "__version__"
                },
                {
                    "type": "promptString",
                    "id": "git-tag-message",
                    "description": "Git tag message.",
                    "default": "New version"
                }
            ]
        }

    **Git hooks example**

    Create folder git_hooks with git hook file - for prec commit name must be `pre-commit`
    (with no extension). Hooks in git folder are gitignored by default (and hooks is not visible
    on first sight).

    Then add hook to git settings - run in terminal (last arg is path (created folder))::

        $ git config core.hooksPath git_hooks

    In created folder on first two lines copy this

    >>> #!/usr/bin/env python
    >>> # -*- coding: UTF-8 -*-

    Then just import any function from here and call with desired params. E.g.

    >>> mypythontools.misc.set_paths()
    >>> mypythontools.utils.run_tests()
    >>> mypythontools.set_version('increment')

"""

import subprocess
import argparse
import importlib
from pathlib import Path
import ast

# Lazy loaded
# from git import Repo


import mylogging

from . import misc
from . import deploy as deploy_module


def push_pipeline(
    tests=True,
    version="increment",
    sphinx_docs=True,
    git_params={
        "commit_message": "New commit",
        "tag": "__version__",
        "tag_mesage": "New version",
    },
    deploy=False,
):
    """Run pipeline for pushing and deploying app. Can run tests, generate rst files for sphinx docs,
    push to github and deploy to pypi. git_params can be configured not only with function params,
    but also from command line with params and therefore callable from terminal and optimal to run
    from IDE (for example with creating simple VS Code task).

    Check utils module docs for implementation example.

    Args:
        tests (bool, optional): Whether run pytest tests. Defaults to True.
        version (str, optional): New version. E.g. '1.2.5'. If 'increment', than it's auto incremented.
            If None, then version is not changed. 'Defaults to "increment".
        sphinx_docs((bool, list)): Whether generate sphinx apidoc and generate rst files for documentation.
            Some files in docs source can be deleted - check `sphinx_docs` docstrings for details and insert
            `exclude_paths` list if have some extra files other than ['conf.py', 'index.rst', '_static', '_templates'].
        git_params (dict, optional): Git message, tag and tag mesage. If empty dict - {},
            than files are not git pushed. If tag is '__version__', than is automatically generated
            from __init__ version. E.g from '1.0.2' to 'v1.0.2'.
            Defaults to { 'commit_message': 'New commit', 'tag': '__version__', 'tag_mesage': 'New version' }.
        deploy (bool, optional): Whether deploy to PYPI. Defaults to False.
    """

    if not all([misc.root_path, misc.app_path, misc.init_path]):
        misc.set_paths()

    if tests:
        run_tests()

    if version:
        set_version(version)

    if isinstance(sphinx_docs, list):
        sphinx_docs_regenerate(exclude_paths=sphinx_docs)
    elif sphinx_docs:
        sphinx_docs_regenerate()

    if git_params:
        parser = argparse.ArgumentParser(
            description="Prediction framework setting via command line parser!"
        )

        parser.add_argument(
            "--commit_message",
            type=str,
            help="Commit message. Defaults to: 'New commit'",
        )
        parser.add_argument(
            "--tag",
            type=str,
            help="Tag. E.g 'v1.1.2'. If '__version__', get the version. Defaults to: '__version__'",
        ),
        parser.add_argument(
            "--tag_mesage", type=str, help="Tag message. Defaults to: 'New version'"
        ),

        parser_args_dict = {
            k: v
            for k, v in parser.parse_known_args()[0].__dict__.items()
            if v is not None
        }

        git_params.update(parser_args_dict)

        git_push(
            commit_message=git_params["commit_message"],
            tag=git_params["tag"],
            tag_message=git_params["tag_mesage"],
        )

    if deploy:
        deploy_module.deploy_to_pypi()


def git_push(commit_message, tag="__version__", tag_message=None):
    """Stage all changes, commit, add tag and push. If tag = '__version__', than tag
    is infered from __init__.py.

    Args:
        commit_message (str): Commit message.
        tag (str): If tag is '__version__', than is automatically generated
            from __init__ version. E.g from '1.0.2' to 'v1.0.2'.
        tag_message (str): [description]
    """

    from git import Repo

    subprocess.run(["git", "add", "."], shell=True, check=True, cwd=misc.root_path)

    subprocess.run(
        ["git", "commit", "-m", commit_message],
        shell=True,
        check=True,
        cwd=misc.root_path,
    )

    if not tag_message:
        tag_message = "New version"

    if tag == "__version__":
        tag = f"v{get_version()}"

    Repo(misc.root_path).create_tag(tag, message=tag_message)

    subprocess.run(
        ["git", "push", "--follow-tags"], shell=True, check=True, cwd=misc.root_path
    )


def set_version(version="increment"):
    """Change your version in your __init__.py file.
    If version is 'increment', it will increment your __version__
    in you __init__.py by one.

    Raises:
        ValueError: If no __version__ is find. Try set init_path...
    """

    with open(misc.init_path, "r") as init_file:

        list_of_lines = init_file.readlines()

        for i, j in enumerate(list_of_lines):
            if j.startswith("__version__"):

                delimiter = '"' if '"' in j else "'"
                delimited = j.split(delimiter)

                if version == "increment":
                    version_list = delimited[1].split(".")
                    version_list[2] = str(int(version_list[2]) + 1)
                    delimited[1] = ".".join(version_list)

                else:
                    delimited[1] = version

                list_of_lines[i] = delimiter.join(delimited)
                break

        else:
            raise ValueError(
                "__version__ variable not found in __init__.py. Try set init_path."
            )

    with open(misc.init_path, "w") as init_file:

        init_file.writelines(list_of_lines)


def get_version():
    """Get version info from __init__.py file.

    Returns:
        str: String of version from __init__.py.

    Raises:
        ValueError: If no __version__ is find. Try set init_path...
    """

    with open(misc.init_path, "r") as init_file:

        for line in init_file:

            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]

        else:
            raise ValueError(
                "__version__ variable not found in __init__.py. Try set init_path."
            )


def run_tests(test_path=None, test_coverage=False):
    import pytest

    """Run tests. If any test fails, raise an error.

    Args:
        test_path ((str, pathlib.Path), optional): Usually autodetected (if root_path / tests).
            Defaults to None.
        test_coverage(bool, optional): Whether run test coverage plugin. If True, pytest-cov must be installed. Defaults to True

    Raises:
        Exception: If any test fail, it will raise exception (git hook do not continue...).
    """

    if not test_path:
        test_path = misc.root_path / "tests"

    if not test_coverage:
        pytest_args = ["-x", test_path.as_posix()]
    else:
        pytest_args = ["-x", "--cov", "./", test_path.as_posix()]

    pytested = pytest.main(pytest_args)

    if pytested == 1:
        raise Exception("Pytest failed")


def sphinx_docs_regenerate(
    docs_path=None, build_locally=0, git_add=True, exclude_paths=[]
):
    """This will generate all rst files necessary for sphinx documentation generation with sphinx-apidoc.
    It automatically delete removed and renamed files.

    Note:
        All the files except ['conf.py', 'index.rst', '_static', '_templates'] will be deleted!!!
        Because if some files would be deleted or renamed, rst would stay and html was generated.
        If you have some extra files or folders in docs source - add it to `exclude_paths` list.

    Function suppose sphinx build and source in separate folders...

    Args:
        docs_path ((str, Path), optional): Where source folder is. Usually infered automatically.
            Defaults to None.
        build_locally (bool, optional): If true, build build folder with html files locally.
            Defaults to 0.
        git_add (bool, optional): Whether to add generated files to stage. False mostly for
            testing reasons.
        exclude_paths ((list)): List of files and folder names that will not be deleted.
            ['conf.py', 'index.rst', '_static', '_templates'] are excluded by default.

    Note:
        Function suppose structure of docs like::

            -- docs
            -- -- source
            -- -- -- conf.py
            -- -- make.bat

        If you are issuing error, try set project root path with `set_root`
    """

    if not importlib.util.find_spec("sphinx"):
        raise ImportError(
            "Sphinx library is necessary for docs generation. Install via `pip install sphinx`"
        )

    if not docs_path:
        if misc.root_path:
            docs_path = misc.root_path / "docs"
        else:
            raise NotADirectoryError(
                mylogging.return_str(
                    "`docs_path` not found. Setup it with parameter `docs_path` or use `misc.set_paths()` function."
                )
            )

    if not all([misc.app_path, misc.root_path]):
        mylogging.return_str("Paths are not known. First run `misc.set_paths()`.")

    docs_source_path = docs_path / "source"

    for p in Path(docs_source_path).iterdir():
        if p.name not in [
            "conf.py",
            "index.rst",
            "_static",
            "_templates",
            *exclude_paths,
        ]:
            p.unlink()

    if build_locally:
        subprocess.run(["make", "html"], shell=True, cwd=docs_path, check=True)

    subprocess.run(
        ["sphinx-apidoc", "-f", "-e", "-o", "source", misc.app_path.as_posix()],
        shell=True,
        cwd=docs_path,
        check=True,
    )

    if git_add:
        subprocess.run(
            ["git", "add", "docs"], shell=True, cwd=misc.root_path, check=True
        )


def generate_readme_from_init(git_add=True):
    """Because i had very similar things in main __init__.py and in readme. It was to maintain news
    in code. For better simplicity i prefer write docs once and then generate. One code, two use cases.

    Why __init__? - Because in IDE on mouseover developers can see help.
    Why README.md? - Good for github.com

    If issuing problems, try misc.set_root() to library path.

    Args:
        git_add (bool, optional): Whether to add generated files to stage. False mostly
            for testing reasons.
    """

    with open(misc.init_path) as fd:
        file_contents = fd.read()
    module = ast.parse(file_contents)
    docstrings = ast.get_docstring(module)

    if docstrings is None:
        docstrings = ""

    with open(misc.root_path / "README.md", "w") as file:
        file.write(docstrings)

    if git_add:
        subprocess.run(
            ["git", "add", "README.md"], shell=True, cwd=misc.root_path, check=True
        )
